package org.etsi.its.configtester.generic;

import java.io.IOException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Observer;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.etsi.its.configtester.IFsapConfigTester;
import org.etsi.its.network.IUdpImplEvent;
import org.etsi.its.network.UdpImpl;
import org.etsi.ttcn.codec.its.configtester.CfPduId;

public class FsapConfigTester extends IFsapConfigTester implements IUdpImplEvent {
    
    /**
     * Module version
     */
    public static final String Version = "1.0.0.0";
    
    /**
     * Logger instance
     */
    private final static Logger _logger = Logger.getLogger("org.etsi.its");
    
    /**
     * Regex pattern for module settings decoding operation
     */
    private Pattern _pattern;
    
    private UdpImpl _cfSendPort;
    
    private boolean _isInitDone;
    
    private Boolean _utResult;
    
    /**
     * Constructor
     */
    public FsapConfigTester() {
        super();
        
        _logger.entering("FsapConfigTester", "Constructor", String.format("version: %s", Version));
        _pattern = Pattern.compile("(\\w+)\\:(\\S+)\\:(\\d+)\\:(\\d+):(.*)"); // For Java regex test, see http://www.cis.upenn.edu/~matuszek/General/RegexTester/regex-tester.html
        _cfSendPort = null;
        _isInitDone = false;
        _utResult = false;
    } // End of Constructor 
    
    @Override
    public synchronized void addObserver(Observer observer) { 
        if (this.countObservers() == 0) { 
            super.addObserver(observer);
        }
    }
    
    @Override
    public boolean initialize(String settings) {
        _logger.entering("FsapConfigTester", this + ".initialize", (settings == null) ? "(null)" : settings);
        
        // Sanity check
        if ((settings == null) || settings.isEmpty()) {
            _logger.severe("FsapConfigTester.initialize: Wrong parameters");
            return false;
        }
        
        try {
            Matcher matcher = _pattern.matcher(settings);
            if (matcher.find()) {
                _cfSendPort = new UdpImpl();
                _cfSendPort.initialize(matcher.group(2), Integer.parseInt(matcher.group(3)), Integer.parseInt(matcher.group(4)), this);
            } else {
                _logger.severe("FsapConfigTester.initialize: Malformed parameters (2)");
                return false;
            }
            
            _cfSendPort.send(new byte[] { (byte) 0xaa } ); // ByteHelper.hexStringToByteArray(matcher.group(5)));
            // A response is expected 
            do {
                try {
                    Thread.sleep(10);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                    break; // Exit 'do-while' loop 
                }
            } while (!_isInitDone); // FIXME Add counter 
            
            _logger.exiting("FsapConfigTester", "initialize", _utResult);
            return _utResult.booleanValue();
        } catch (UnknownHostException e) {
            e.printStackTrace();
        } catch (SocketException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return false;
    }
    
    @Override
    public boolean uninitialize() {
        _logger.entering("FsapConfigTester", "uninitialize");
        
        if (this.countObservers() != 0) { 
            super.deleteObservers();
        }
        if (_cfSendPort != null) {
            _cfSendPort.uninitialize();
            _cfSendPort = null;
        }
        
        return true;
    }
    
    public void dispose() {
        _logger.entering("FsapConfigTester", "dispose");
        
        uninitialize();
    }
    
    @Override
    public void onReceive(byte[] packet) {
        _logger.entering("FsapConfigTester", "onReceive");
        
        if (packet[0] == 0x59) { // CfResult received 
            _logger.info("CfResult received");
            if ((packet.length == 4) && (packet[0] == CfPduId.value("CfResult"))) { 
                _utResult = new Boolean((boolean)(packet[3] == 0x01));
            } else { 
                _utResult = new Boolean(false);
            }
            _isInitDone = true; 
        } else {
            // Notify ConfigTester 
            _logger.info("Notify ConfigTester");
            setChanged(); 
            notifyObservers(packet);
        }
        _logger.exiting("FsapConfigTester", "onReceive");
    }
    
} // End of class FsapConfigTester 
