package org.etsi.its.adapter.stf455.dispatcher;

import java.io.IOException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.etsi.common.ByteHelper;
import org.etsi.its.network.IUdpImpl;
import org.etsi.its.network.IUdpImplEvent;
import org.etsi.its.network.UdpImpl;
//import java.util.List;

public class DispatcherTrigger extends Thread implements IUdpImplEvent {
    
    /**
     * Module version
     * version 1.0.1.0
     *         new control variable _Asn1Version
     *           1 indicates version valid until SEP-2013
     *           2 indicates version, published in SEP/OCT-2013
     *              
     * version 1.0.2.0
     *      changed behavior to match ATSP testing requirements
     *      for details see webex meeting 2013-DEC-17
     *              
     * version 1.0.3.0
     *      changed behavior to match testing requirements
     *      for details see webex meeting 2014-JAN-16
     *         
     */
    public static final String Version = "1.0.4.3";
    private static final int _Asn1Version = 2;
    
    private static final byte UtResultTag = (byte) 0x24; // See org.etsi.ttcn.codec.its.uppertester.UtPduId
    private static final byte UtFntpInitializeTag = (byte) 0x90;
    private static final byte UtFntpEventIndTag = (byte) 0x9a;
    private static final byte UtFsapInitializeTag = (byte) 0xa0;
    private static final byte UtFsapEventIndTag = (byte) 0xa2; 
    private static final byte UtIicpInitializeTag = (byte) 0xb0;
    private static final byte UtIicpEventTag_mCmdRs = (byte) 0xb1; 
    private static final byte UtIicpEventTag_mf_Request_request = (byte) 0xb2; 
    private static final byte UtIicpEventTag_mn_Request_request = (byte) 0xb3; 
    private static final byte UtIicpEventTag_mi_Request_request = (byte) 0xb4; 
    private static final byte UtIicpEventTag_mi_Get_request = (byte) 0xb5; 
    private static final byte UtIicpEventTag_mi_Set_request = (byte) 0xb6; 
    private static final byte UtIicpEventTag_mf_Command_confirm = (byte) 0xb7; 
    private static final byte UtIicpEventTag_mn_Command_confirm = (byte) 0xb8; 
    private static final byte UtIicpEventTag_mi_Command_confirm = (byte) 0xb9; 
    private static final byte UtIicpEvent_mi_Get_confirm = (byte) 0xba;
    private static final byte UtIicpEvent_mi_Set_confirm = (byte) 0xbb;
    private static final byte UtIicpEventIndTag_McmdRq = (byte) 0xc0; 
    private static final byte UtIicpEventIndTag_mf_Command_request = (byte) 0xc1; 
    private static final byte UtIicpEventIndTag_mn_Command_request = (byte) 0xc2; 
    private static final byte UtIicpEventIndTag_mi_Command_request = (byte) 0xc3; 
    private static final byte UtAtspEventIndTag = (byte) 0xff; // FIXME To be changed
    private static final byte CfFntpInitializeTag = (byte) 0x10;
    private static final byte CfFsapInitializeTag = (byte) 0x20;
    private static final byte CfIicpInitializeTag = (byte) 0x30;
    private static final byte CfAtspInitializeTag = (byte) 0x40;
    private static final byte CfResultTag = (byte) 0x25;
    
    /**
     * Logger instance
     */
    private final static Logger _logger = Logger.getLogger("org.etsi.its.adapter.stf455.Dispatcher");

    
    /**
     * UT module active or not
     */
    private static Boolean UtActive = false;
    
    /**
     * LT module active or not
     */
    private static Boolean LtActive = false;
    
    /**
     * CF module active or not
     */
    private static Boolean CfActive = false;
    
    /**
     * using simulator instead of communicating to a SUT
     */
    private static Boolean IicpSimulate = false;
    
        
    /**
     * Flag used to terminate infinite loop in internal thread
     */
    private boolean _running;
    
    private DispatcherTriggerUt ut;
    private DispatcherTriggerLt lt;
    private DispatcherTriggerCf cf;
    
    /**
     * Dispatcher UDP module TTE side
     */
    private IUdpImpl _udpUt;
    private IUdpImpl _udpLt;
    private IUdpImpl _udpCf;
    
    /**
     * Dispatcher UDP module IICP port
     */
    private static IUdpImpl _udpIicp;
    
    /**
     * port settings for IICP messages
     */
    private    static    int TestSystemScuId = 3;
    private    static    int SutScuId = 12;            // for testing @ Commsignia SUT
    
    /**
     * data type identifier to be used in IICP messages forwarded from UT, LT, CF
     * will be modified when receiving the TestConfigIICP message (0xFF 0x01 0xnn)
     * zero values prevent messages to be sent to IICP port 
     */
    private static byte UtDataTypeId = 0;
    private static byte LtDataTypeId = 0;
    private static byte CfDataTypeId = 0;

    /**
     * controls if de-/encapsulation is to be performed
     */
    
    private static boolean IicpTesting = false;

    /**
     * only used for IICP testing
     */
    private static byte lastSimUtOutRqContentId = (byte)254;    // contentId used in last received SimUtOutRq
    private static byte lastLtReqRef = 0;                        // reqRef from last processed IIC-Request received on LT port    
    
    /**
     * pduCounter for IICP messages
     */
    private    int pduCounterIicp = 0;
    
    /**
     * Note: When using List<byte[]> then keep in mind the following!
     * don't use values of 128 or higher, this will not work properly !!!!
     */
    private static final Integer _MaxBufferSize = 0x7FFF;
    
    /**
     * enable or disable output of additional information using System.out 
     */
    private static final boolean debugBuffer = true;
    
    private Queue<byte[]> _recvUtPacketBuffer;
    private Queue<byte[]> _recvLtPacketBuffer;
    private Queue<byte[]> _recvCfPacketBuffer;
    private Queue<byte[]> _recvIicpPacketBuffer;
    
    private Integer _recvUtPacketIdx;
    private Integer _recvLtPacketIdx;
    private Integer _recvCfPacketIdx;
    private Integer _recvIicpPacketIdx;
    
    private Integer _processedUtPacketIdx;
    private Integer _processedLtPacketIdx;
    private Integer _processedCfPacketIdx;
    private Integer _processedIicpPacketIdx;
    
    public DispatcherTrigger(final String utSettings, 
            final String ltSettings, 
            final String cfSettings, 
            final String iicpSettings,
            final int     sutItsScuIdSetting) {
        _logger.setLevel(Level.ALL);
        _logger.entering(this.toString(), "dispatcherTrigger, Version " + Version, String.format("%s:%s:%s:%s:sutItsScuId=%d", utSettings, ltSettings, cfSettings, iicpSettings, sutItsScuIdSetting));
        
        try {
            ut = new DispatcherTriggerUt(utSettings);
            lt = new DispatcherTriggerLt(ltSettings);
            cf = new DispatcherTriggerCf(cfSettings);

            Pattern pattern = Pattern.compile("(\\w+)\\:(\\S+)\\:(\\d+)\\:(\\d+)"); // For Java regex test, see http://www.cis.upenn.edu/~matuszek/General/RegexTester/regex-tester.html
            Matcher matcher = pattern.matcher(iicpSettings);
            if (matcher.find()) {
                if ((Integer.parseInt(matcher.group(3)) > 0) && (Integer.parseInt(matcher.group(4)) > 0)) {
                    _recvIicpPacketBuffer = new LinkedList<byte[]>();
                    _recvIicpPacketIdx = 0;
                    _processedIicpPacketIdx = 0;
                    _udpIicp = new UdpImpl();
                    _udpIicp.initialize(matcher.group(2), Integer.parseInt(matcher.group(3)), Integer.parseInt(matcher.group(4)), this);
                } else {
                    IicpSimulate = true;
                    _logger.severe("IICP simuation");
                }
            } else {
                _logger.severe("Malformed parameters (5)");
                throw new RuntimeException("Malformed parameters (5)");
            }
            SutScuId = sutItsScuIdSetting;
            
            _logger.exiting(this.toString(), "dispatcherTrigger");
        } catch (UnknownHostException e) {
            e.printStackTrace();
        } catch (SocketException e) {
            e.printStackTrace();
        }
    } // End of Constructor

    @Override
    public void run() { 
        boolean busy;
        
        _logger.setLevel(Level.ALL);    // something goes wrong with logging, some messages disappear
        _logger.entering(this.toString(), "run");
        
        _running = true;
        while (_running) {
            try {
                Thread.sleep(5);    // 100
                busy = true;
                while (busy) {
                    busy = false;
                    if (UtActive) {
                        //if (_processedUtPacketIdx != _recvUtPacketIdx) { // There is at least one packet to be processed
                        if (!_recvUtPacketBuffer.isEmpty()) {
                            busy = true;
                            if (debugBuffer) System.out.println("@@@@@ Ut.proc processedIdx=" + _processedUtPacketIdx + " - recvIdx=" + _recvUtPacketIdx + " - length=" + _recvUtPacketBuffer.size());
                            // Retrieve packet 
                            synchronized(_recvUtPacketBuffer) {
                                try {
                                    // get & remove item from list 
                                    byte[] packet = _recvUtPacketBuffer.remove();
                                    _processedUtPacketIdx++;
                                    if (_processedUtPacketIdx > _MaxBufferSize) _processedUtPacketIdx = 0;
                                    // Process it
                                    _logger.info(String.format("Process UT message id 0x%02x", packet[0]));
                                    // process UT messages
                                    ut.DispatchMessage(packet);
                                } catch (NoSuchElementException e) {
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
    
                    if (LtActive) {
                        //if (_processedLtPacketIdx != _recvLtPacketIdx) { // There is at least one packet to be processed
                        if (!_recvLtPacketBuffer.isEmpty()) {
                            busy = true;
                            if (debugBuffer) System.out.println("@@@@@ Lt.proc processedIdx=" + _processedLtPacketIdx + " - recvIdx=" + _recvLtPacketIdx + " - length=" + _recvLtPacketBuffer.size());
                            // Retrieve packet 
                            synchronized(_recvLtPacketBuffer) {
                                try {
                                    // get & remove item from list 
                                    byte[] packet = _recvLtPacketBuffer.remove();
                                    _processedLtPacketIdx++;
                                    if (_processedLtPacketIdx > _MaxBufferSize) _processedLtPacketIdx = 0;
                                    // Process it
                                    _logger.info(String.format("Process LT message id 0x%02x", packet[0]));
                                    // process LT messages
                                    lt.DispatchMessage(packet);
                                } catch (NoSuchElementException e) {
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
                    
                    if (CfActive) {
                        //if (_processedCfPacketIdx != _recvCfPacketIdx) { // There is at least one packet to be processed
                        if (!_recvCfPacketBuffer.isEmpty()) {
                            busy = true;
                            if (debugBuffer) System.out.println("@@@@@ Cf.proc processedIdx=" + _processedCfPacketIdx + " - recvIdx=" + _recvCfPacketIdx + " - length=" + _recvCfPacketBuffer.size());
                            // Retrieve packet 
                            synchronized(_recvCfPacketBuffer) {
                                try {
                                    // get & remove item from list 
                                    byte[] packet = _recvCfPacketBuffer.remove();
                                    _processedCfPacketIdx++;
                                    if (_processedCfPacketIdx > _MaxBufferSize) _processedCfPacketIdx = 0;
                                    // Process it
                                    _logger.info(String.format("Process CF message id 0x%02x", packet[0]));
                                    // process CF messages
                                    cf.DispatchMessage(packet);
                                } catch (NoSuchElementException e) {
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
                    
                    if (!IicpSimulate) {
                        //if (_processedIicpPacketIdx != _recvIicpPacketIdx) { // There is at least one packet to be processed
                        if (!_recvIicpPacketBuffer.isEmpty()) {
                            busy = true;
                            if (debugBuffer) System.out.println("@@@@@ Iicp.proc processedIdx=" + _processedIicpPacketIdx + " - recvIdx=" + _recvIicpPacketIdx + " - length=" + _recvIicpPacketBuffer.size());
                            // Retrieve packet
                            synchronized(_recvIicpPacketBuffer) {
                                try {
                                    // get & remove item from list 
                                    byte[] packet = _recvIicpPacketBuffer.remove();
                                    _processedIicpPacketIdx++;
                                    if (_processedIicpPacketIdx > _MaxBufferSize) _processedIicpPacketIdx = 0;
                                    // Process it
                                    _logger.info(String.format("Process IICP message id 0x%02x", packet[0]));
                                    // process IICP messages
                                    DispatchMessage(packet);
                                } catch (NoSuchElementException e) {
                                    e.printStackTrace();
                                }
                            }
                        }
                    }
                }    // End of while (busy)
            } catch (InterruptedException /*| IOException*/ e) {
                e.printStackTrace();
            }
        } // End of 'while (_running)' statement
        
        _logger.exiting(this.toString(), "run");
    } // End of method run
    
    /**
     * Terminate internal thread
     */
    public void terminate() {
        _logger.entering(this.toString(), "terminating");
        _running = false;
        try {
            join();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        
        if (UtActive) {_udpUt.uninitialize(); _recvUtPacketBuffer.clear();}
        if (LtActive) {_udpLt.uninitialize(); _recvLtPacketBuffer.clear();}
        if (CfActive) {_udpCf.uninitialize(); _recvCfPacketBuffer.clear();}
        if (!IicpSimulate) {_udpIicp.uninitialize(); _recvIicpPacketBuffer.clear();}
        
        _recvUtPacketIdx = 0;
        _recvLtPacketIdx = 0;
        _recvCfPacketIdx = 0;
        _recvIicpPacketIdx = 0;
        _processedUtPacketIdx = 0;
        _processedLtPacketIdx = 0;
        _processedCfPacketIdx = 0;
        _processedIicpPacketIdx = 0;
        
        _logger.exiting(this.toString(), "terminated");
    } // End of method terminate
    
    @Override
    public void onReceive(byte[] packet) {
        ByteHelper.dump("DispatchTriggerIicp.onReceive:", packet);
        synchronized(_recvIicpPacketBuffer) {
            try {
                _recvIicpPacketBuffer.add(packet);    
            } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                e.printStackTrace();
            }
            _recvIicpPacketIdx++;
            if (_recvIicpPacketIdx > _MaxBufferSize) _recvIicpPacketIdx = 0;
        };
        if (debugBuffer) System.out.println("@@@@@ Iicp.rx processedIdx=" + _processedIicpPacketIdx + " - recvIdx=" + _recvIicpPacketIdx + " - length=" + _recvIicpPacketBuffer.size());
    }

    // debug helper functions
    public void InjectMsgUtPort(byte[] msg) {
        synchronized(_recvUtPacketBuffer) {
            try {
                _recvUtPacketBuffer.add(msg);    
            } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                e.printStackTrace();
            }
            _recvUtPacketIdx++;
            if (_recvUtPacketIdx > _MaxBufferSize) _recvUtPacketIdx = 0;
        }
    }
    
    public void InjectMsgLtPort(byte[] msg) {
        synchronized(_recvLtPacketBuffer) {
            try {
                _recvLtPacketBuffer.add(msg);    
            } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                e.printStackTrace();
            }
            _recvLtPacketIdx++;
            if (_recvLtPacketIdx > _MaxBufferSize) _recvLtPacketIdx = 0;
        }
    }
    
    public void initDebugIicp() {
        UtDataTypeId = (byte)240;
        LtDataTypeId = (byte)241;
        IicpTesting  = true;
    }
    // end of debug helper functions
    
    /**
     * retrieve contentId for next SimUtInRequest() based on stored data from previous message exchanges
     * @return new contentId
     */
    public byte GetContentId () {
        byte    contentId = 0;

        /**
         * first variant, based on last SimUtOutRequest data
         */
        if (lastSimUtOutRqContentId < 250) {
            contentId = (byte)(lastSimUtOutRqContentId + 1);    // contentId used in last received SimUtOutRq
        } else {
        /**
         * second, alternative variant, based on mapping table between RefIICPReq/Res & contentID
         * not needed, normally the first should work
         */
        // lastLtReqRef = 0;                // reqRef from last processed IIC-Request received on LT port
        //if ((byte)9 == lastLtReqRef) contentId = 20;
        
        }
        return(contentId);
    }
    
    /**
     * removes the IICP frame and forwards the content to the UT port
     * @param msg    - IICP message
     * @param offs    - starting index of contained message
     * @param len   - length of embedded data field
     */
    public void ForwardMessageToUtPort(byte[] msg, int offs, int len, byte utEventTag)
    {
        _logger.info("FwdMsgToUtPort, offs = " + Integer.toString(offs) +  ", length = " + Integer.toString(len));
        int size=len;
        byte [] tx = new byte[size + 1];
        
        tx[0] = utEventTag; 
        System.arraycopy(msg, offs, tx, 1, size);    // copy internal message frame
        ByteHelper.dump("Iicp.Forward msg to UT:", tx);
        try {
            _udpUt.send(tx);
        } catch (/*InterruptedException |*/ IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * removes the IICP frame and forwards the content to the LT port
     * @param msg    - IICP message
     * @param offs    - starting index of contained message
     * @param len   - length of embedded data field
     */
    public void ForwardMessageToLtPort(byte[] msg, int offs, int len)
    {
        _logger.info("FwdMsgToLtPort, offs = " + Integer.toString(offs) +  ", length = " + Integer.toString(len));
        int size=len;
        byte [] tx = new byte[size];
        System.arraycopy(msg, offs, tx, 0, size);    // copy internal message frame
        ByteHelper.dump("Iicp.Forward msg to LT:", tx);
        try {
            _udpLt.send(tx);
        } catch (/*InterruptedException |*/ IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * forwards the content to the CF port, unchanged
     * @param msg    - IICP message
     * @param offs    - starting index of contained message
     * @param len   - length of embedded data field
     */
    public void ForwardMessageToCfPort(byte[] msg, int offs, int len)
    {
        _logger.info("FwdMsgToCfPort, offs = " + Integer.toString(offs) +  ", length = " + Integer.toString(len));
        int size=len;
        byte [] tx = new byte[size];
            System.arraycopy(msg, offs, tx, 0, size);    // copy internal message frame
        ByteHelper.dump("Iicp.Forward msg to CF:", tx);
        try {
            _udpCf.send(tx);
        } catch (/*InterruptedException |*/ IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * analyze and dispatch incoming IICP messages 
     * @param msg - received message
     */
    public void DispatchMessage(byte[] msg)
    {
        boolean msgHandled = false;
        // check for minimum length of the message
        if (msg.length < 12) {
            _logger.info("Iicp.DispatchMessage, too short, length = " + Integer.toString(msg.length));
            return;
        }
        
        // select some identifier information from incoming message
        int i = 6;
        int offs = 0;    // starting index of the embedded message
        byte pduRef = -1;
        int  pduLength = -1;
        byte pduResRef = -1;
        int  srcScuId   = (((int)msg[0] & 0xFF) << 8) + ((int)msg[1] & 0xFF);
        int  destScuId  = (((int)msg[2] & 0xFF) << 8) + ((int)msg[3] & 0xFF);
        int  pduCounter = (((int)msg[4] & 0xFF) << 8) + ((int)msg[5] & 0xFF);
        pduRef     = msg[i++];    // only 0 and 1 expected
        if (1 == _Asn1Version) {
            if ((msg[i] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                pduLength  = (((int)msg[i++] & 0x3F) << 8) + ((int)msg[i++]  & 0xFF);
            } else {
                pduLength  = (int)msg[i++];    
            }
            pduResRef  = msg[i++];    // IICP data type identifier
        } else if (2 == _Asn1Version) {
            pduResRef  = msg[i++];    // IICP data type identifier
            if ((msg[i] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                pduLength  = (((int)msg[i++] & 0x3F) << 8) + ((int)msg[i++]  & 0xFF);
            } else {
                pduLength  = (int)msg[i++];    
            }
        }
        offs = i;
        int  pduIicpResLen = pduLength;
        if (_Asn1Version == 1) { // TODO YGA Not sure what to do with this code in case of _Asn1Version == 2
            // valid only for pduResRef = {242 .. 255}
            if ((msg[i] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                pduIicpResLen  = (((int)msg[i++] & 0x3F) << 8) + ((int)msg[i++]  & 0xFF);
            } else {
                pduIicpResLen  = (int)msg[i++];    
            }
            //pduIicpResLen  = (int)msg[i++];
            
            // the following line causes errors during tests, there was one missing leading byte, or not - ??? 
             offs = i;    // points to the first element of the data content, the length information must be forwarded too
             //byte pduIicpRes = msg[i++];    // only zero value expected
        }
        // check for destination ITS-scuId
        if (destScuId != TestSystemScuId) {
            if (srcScuId == SutScuId) {    // for IICP testing
                try {
                    _udpLt.send(msg);
                } catch (/*InterruptedException |*/ IOException e) {
                    e.printStackTrace();
                }
                return;
            } else {    // skip if not mine
            _logger.info("Iicp.DispatchMessage, skipped, destScuId = " + Integer.toString(destScuId));
            return;
        }
        }
        // check for sourceITS-scuId, skip if not the SUT
        if ((srcScuId != SutScuId) && !IicpTesting) {
            _logger.info("Iicp.DispatchMessage, skipped, srcScuId = " + Integer.toString(srcScuId));
            return;
        }
            
        // check pduCounter -->> ??
        
        // check pduRef
        if (pduRef > 1) {
            _logger.info("Iicp.DispatchMessage, skipped, pduRef = " + Integer.toString((int)pduRef & 0xFF));
            return;
        }
        
        // check for messages to be forwarded
        if ((pduResRef == (byte)240) || (pduResRef == (byte)241)) {            // simUtIn/simUtOut
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                if (IicpTesting && (pduResRef == (byte)241)) {    // only for SimUtOutReq
                    int k=8;
                    if ((msg[k] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                        k++;
                    }
                    k++;
                    lastSimUtOutRqContentId = msg[k];    // contentId used in last received SimUtOutRq
                }
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                byte utEventIndTag = (byte) 0xff;
                if (msg[9] == (byte)20) {
                    utEventIndTag = UtIicpEventIndTag_McmdRq;
                } else if (msg[9] == (byte)0){
                    utEventIndTag = UtIicpEventIndTag_mf_Command_request;
                } else if (msg[9] == (byte)2){
                    utEventIndTag = UtIicpEventIndTag_mn_Command_request;
                } else if (msg[9] == (byte)4){
                    utEventIndTag = UtIicpEventIndTag_mi_Command_request;
                } else {
                    utEventIndTag = (byte) 0xff;
                }
                ForwardMessageToUtPort(msg, offs + 2, pduIicpResLen - 2, utEventIndTag);
            }            
        } else if ((pduResRef == (byte)242) || (pduResRef == (byte)243)) {            // simAutCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToUtPort(msg, offs, pduIicpResLen, UtAtspEventIndTag);
            }            
        } else if ((pduResRef == (byte)244) || (pduResRef == (byte)245)) {    // simFutCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToUtPort(msg, offs, pduIicpResLen, UtFsapEventIndTag);
            }
        } else if ((pduResRef == (byte)246) || (pduResRef == (byte)247)) {    // simNutCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToUtPort(msg, offs, pduIicpResLen, UtFntpEventIndTag);
            }            
        } else if ((pduResRef == (byte)248) || (pduResRef == (byte)249)) {    // simIutCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToUtPort(msg, offs, pduIicpResLen, (byte) 0xff); // FIXME To be changed
            }            
        } else if ((pduResRef == (byte)250) || (pduResRef == (byte)251)) {    // simAltCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToLtPort(msg, offs, pduIicpResLen);
            }            
        } else if ((pduResRef == (byte)252) || (pduResRef == (byte)253)) {    // simFltCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToLtPort(msg, offs, pduIicpResLen);
            }            
        } else if ((pduResRef == (byte)254) || (pduResRef == (byte)255)) {    // simNltCmdRq/ReqRq
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                ForwardMessageToLtPort(msg, offs, pduIicpResLen);
            }            
        //} else if (pduResRef == 0) {                            // aliveRq
            // TODO answer to alive request
            
        } else if (pduResRef == 0) {                            // alive
            // monitor for alive(new) messages
            // - remember freshly assigned ITS-SCU-ID of the SUT/IUT
            // - response to the request if the ID is in the allowed range
            
            // forward the message to the LT port
//            ForwardMessageToLtPort(msg, offs, pduIicpResLen);
            ForwardMessageToLtPort(msg, 0, msg.length);
        } else if (pduResRef == 9) {                            // McmdRq
            // TODO only if upper tester access is active
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO check with Yann / HJF
                
                // check for IICP/ConfigTestIICP response
                if ((msg[offs] == (byte)0xFF) && 
                        //((msg[offs+1] == (byte)0x03) || (msg[offs+1] == (byte)0x01)) &&
                        // HJF: only value 0x01 is allowed, 2014-01-16 meeting
                        (msg[offs+1] == (byte)0x01) &&
                        (msg[offs+2] == (byte)0x00) && (msg[offs+3] == (byte)0x00)) {
                    msgHandled = true;
                    try {
                        _udpUt.send(new byte[] {    // send UtResult to upper tester port
                                (byte)UtResultTag,
                                (byte)0x01
                        });
                    } catch (/*InterruptedException |*/ IOException e) {
                        e.printStackTrace();
                    }
                }
                if (!IicpTesting) msgHandled = true;    // avoid forwarding of response messages
            } else {
                // TODO
            }
            if (!msgHandled) {    // forward message to LT or CF port
                if (IicpTesting) {
                    //ForwardMessageToLtPort(msg, offs, pduIicpResLen); // message body without IICP frame
                    ForwardMessageToLtPort(msg, 0, msg.length);        // whole message
                } else {
                    if (CfActive) {        // forward it to CF port
                        //SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                        ForwardMessageToCfPort(msg, offs, pduIicpResLen);
                    }
                }
            }
        } else  {                        // all other
            if (pduRef == 1) {    // response to a request
                _logger.info("Iicp.DispatchMessage, rx IIC-Response, pduResRef = " + Integer.toString((int)pduResRef & 0xFF) +  ", pduCounter = " + Integer.toString(pduCounter));
                // TODO clear pending waiting for answer
                return;
            } else {
                if (IicpTesting) {    // special handling for selected messages
                    if (((byte)0x02==pduResRef) || // MF-Request-request
                        ((byte)0x04==pduResRef) || // MN-Request-request
                        ((byte)0x06==pduResRef) || // MI-Request-request
                        ((byte)0x0B==pduResRef) || // MA-Request-request
                        ((byte)0x0D==pduResRef)    // MS-Request-request
                    ){
                        i = 0;
                        i+= 8;    // skip src, dest, cnt, reqID, reqRef 
                        if ((msg[i] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                            i+= 2;
                        } else {
                            i++;
                        };
                        i+= 2;    // skip commandRef
                        i++;    // skip ref
                        if ((msg[i] & (byte)0xC0) == (byte)0x80) { // check for length > 127
                            pduIicpResLen  = (((int)msg[i++] & 0x3F) << 8) + ((int)msg[i++]  & 0xFF);
                        } else {
                            pduIicpResLen  = (int)msg[i++];    
                        }
                        offs = i;
                        if ((byte)0x02==pduResRef) { // FIXME To be change according to the message type
                            //ForwardMessageToUtPort(msg, offs, pduIicpResLen, (byte) UtIicpEventIndTag_mcmdRq=0xba);
                        }
                        ForwardMessageToUtPort(msg, offs, pduIicpResLen, (byte) UtIicpEventIndTag_McmdRq);
                    } else {        // keep the old (previous) behavior for all others
                        if (CfActive) {        // forward it to CF port
                            //SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                            ForwardMessageToCfPort(msg, offs, pduIicpResLen);
                        }
                    }
                } else {
                    if (CfActive) {        // forward it to CF port
                        //SendIicpPduResponseAck(TestSystemScuId, SutScuId, pduCounter+1, pduResRef); 
                        ForwardMessageToCfPort(msg, offs, pduIicpResLen);
                    }
                }
            }
            return;
        }
    }     // End of Dispatch Message IICP

    /**
     * Wrap the pay load into an IICP message frame and send it via IICP port to the SUT 
     * @param sourceScuId    - elements of the IICP frame
     * @param destScuId
     * @param pduCounter
     * @param reqRef        
     * @param contentId     - used only for SimUtInRq & SimUtOutRq
     * @param payload       - pay load to be embedded into IICP message
     */
    
    public static void SendIicpPduRequest(int sourceScuId, int destScuId, int pduCounter, byte reqRef, byte contentId, byte[] payload)
    {
        if (reqRef == 0) {    // no valid IICP data type identifier assigned, skip message
            _logger.info("Iicp.SendIicpPduRequest, msg skipped, reqRef = " + Integer.toString(reqRef));
            return;
        }
        int i;
        if ((1 == _Asn1Version) && (reqRef != (byte)240)) {
            i = 10 + payload.length + 2;
        } else /*if (2 == _Asn1Version)*/ {
            if (((byte)240 == reqRef) || ((byte)241 == reqRef)) {    // SimUtIn & SimUtOut do not contain doubled length fields
                i = 10 + payload.length + 3;    // difference in seqRq
            } else {
                i = 10 + payload.length + 3;    // difference in seqRq
            }
        }
            
        if (((byte)240 == reqRef) || ((byte)241 == reqRef)) {    // SimUtIn & SimUtOut do not contain doubled length fields
            if ((payload.length + 2) > 127) i++;    // check for additional bytes in length indicators
        } else {
            if ((payload.length + 4) > 127) i++;    // check for additional bytes in length indicators
            if ((payload.length + 2) > 127) i++;
        }
        
        byte[] msg = new byte[i];
        i = 0;
        msg[i++] = (byte)(sourceScuId >>> 8);        // sourceITS-scuId, 2 bytes
        msg[i++] = (byte)(sourceScuId);            
        msg[i++] = (byte)(destScuId >>> 8);        // destinationITS-scuId, 2 bytes
        msg[i++] = (byte)(destScuId);            
        msg[i++] = (byte)(pduCounter >>> 8);        // pduCounter, 2 bytes, even values only
        msg[i++] = (byte)(pduCounter);             
        msg[i++] = (byte)0x00;                    // pduRequest

        int len;
        if ((1 == _Asn1Version) && (reqRef != (byte)240)) {
            len = payload.length + 2;            // length of the payload + 2
            if ((payload.length + 2) > 127) {        // inserts one additional byte for extended length indicator
                len++;
            }
            if (len > 127) {
                msg[i++] = (byte)(((byte)(len >>> 8)) | ((byte)0x80));
                msg[i++] = (byte)len;
            } else {
                msg[i++] = (byte)len;
            }
            msg[i++] = (byte)reqRef;                // reqReference
        } else /*if (2 == _Asn1Version)*/ {
            msg[i++] = (byte)reqRef;                // reqReference
            if (((byte)240 != reqRef) && ((byte)241 != reqRef)) {    // SimUtIn & SimUtOut do not contain doubled length fields 
                len = payload.length + 1;            // length of the payload + 1
                if ((payload.length + 1) > 127) {        // inserts one additional byte for extended length indicator
                    len++;
                }
                if (len > 127) {
                    msg[i++] = (byte)(((byte)(len >>> 8)) | ((byte)0x80));
                    msg[i++] = (byte)len;
                } else {
                    msg[i++] = (byte)len;
                }
            }
        }
        
        len = payload.length;                    // length of the payload
        if (((byte)240 == reqRef) || ((byte)241 == reqRef)) {    // SimUtIn & SimUtOut do not contain doubled length fields
            len++;        // contentId
        }
        if (len > 127) {
            msg[i++] = (byte)(((byte)(len >>> 8)) | ((byte)0x80));
            msg[i++] = (byte)len;
        } else {
            msg[i++] = (byte)len;
        }
        if (((byte)240 == reqRef) || ((byte)241 == reqRef)) {    // SimUtIn & SimUtOut do not contain doubled length fields
            msg[i++] = (byte)contentId;
        }
        System.arraycopy(payload, 0, msg, i, payload.length);
        i += payload.length;
                        // add separate security data (placeholder)
        if ((1 == _Asn1Version) && (reqRef != (byte)240)) {
            msg[i++] = (byte)0x00;    // seqRq
            msg[i++] = (byte)0x00;    // seqRq
        } else /*if (2 == _Asn1Version)*/ {
            msg[i++] = (byte)0x00;    // noSecurity
            msg[i++] = (byte)0x01;    // len
            msg[i++] = (byte)0x00;    // NULL
        }            
        ByteHelper.dump("SendIicpPduRequest:", msg);
        try {
            _udpIicp.send(msg);
        } catch (/*InterruptedException |*/ IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Generate an IICP Response with ErrorStatus=0 and send it to the IUT
     * @param sourceScuId    - elements of the IICP frame
     * @param destScuId
     * @param pduCounter    - number derived from the request message to be acknowledged (+1)
     * @param reqRef        - resRef taken from the request message
     */
    
    public static void SendIicpPduResponseAck(int sourceScuId, int destScuId, int pduCounter, byte reqRef)
    {
        int i = 15;
        byte[] msg = new byte[i];
        i = 0;
        msg[i++] = (byte)(sourceScuId >>> 8);        // sourceITS-scuId, 2 bytes
        msg[i++] = (byte)(sourceScuId);            
        msg[i++] = (byte)(destScuId >>> 8);            // destinationITS-scuId, 2 bytes
        msg[i++] = (byte)(destScuId);            
        msg[i++] = (byte)(pduCounter >>> 8);        // pduCounter, 2 bytes, even values only
        msg[i++] = (byte)(pduCounter);             
        msg[i++] = (byte)0x01;                        // pduResponse
        if (1 == _Asn1Version) {
            msg[i++] = (byte)0x03;                        // length 
            msg[i++] = (byte)reqRef;                      // reqReference
        } else if (2 == _Asn1Version) {
            msg[i++] = (byte)reqRef;                      // reqReference
            msg[i++] = (byte)0x02;                        // length 
        }
        msg[i++] = (byte)0x01;                        // length 
        msg[i++] = (byte)0x00;                        // positive result
        msg[i++] = (byte)0x00;                        // error status 
        if (1 == _Asn1Version) {
            msg[i++] = (byte)0x00;                        // seqRq
            msg[i++] = (byte)0x00;                        // seqRq
        } else if (2 == _Asn1Version) {
            msg[i++] = (byte)0x00;    // noSecurity
            msg[i++] = (byte)0x01;    // len
            msg[i++] = (byte)0x00;    // NULL
        }            
        ByteHelper.dump("SendIicpPduResponseAck:", msg);
        try {
            _udpIicp.send(msg);
        } catch (/*InterruptedException |*/ IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * set valid IICP data type identifiers to be used for different ports
     * @param TestConfigIICP - provided from the TA to setup the SUT into correct test mode
     */
    public static void SetDataTypeIdentifiersForPorts(byte TestConfigIICP)
    {
        /*    IICP data type identifier
         *       0    ITS-SCUalive
         *      ...
         *       9 McmdRq
         *      ...
         *      13 
         * 
         *     240    SimUtInReq
         *     241    SimUtOutReq     
         *     242    SimAUTcmd
         *     243    SimAUTreq
         *     244    SimFUTcmd
         *     245    SimFUTreq
         *     246    SimNUTcmd
         *     247    SimNUTreq
         *     248    SimIUTcmd
         *     249    SimIUTreq
         *     250    SimALTcmd
         *     251    SimALTreq
         *     252    SimFLTcmd
         *     253    SimFLTreq
         *     254    SimNLTcmd
         *     255    SimNLTreq
         */
        // TODO
        UtDataTypeId = (byte)244;    // preset default values
        LtDataTypeId = (byte)0;
        CfDataTypeId = (byte)9;
        IicpTesting  = false;
        
        // check for "lower layers are used"
        if ((TestConfigIICP & (byte)0x01) == (byte)0x01) IicpTesting  = true; // FIXME Incorrect fix because Iicp will be used for FNTP/FSAP too
//        if ((TestConfigIICP & (byte)0x02) == (byte)0x02) IicpTesting  = true;
        if ((TestConfigIICP & (byte)0x80) == (byte)0x80) { 
            // it implies "lower layers are used"
            IicpTesting  = true;
        }
        
        switch (TestConfigIICP & (byte)0xFC) {
        case (byte)0x80:    // UT: access layer, LT: use physical layer
            UtDataTypeId = (byte)248;
            break;
        case (byte)0x40:    // UT: network&transport layer, LT: use physical layer
            UtDataTypeId = (byte)246;
            LtDataTypeId = (byte)254;
            break;
        case (byte)0x20:    // UT: facilities layer, LT: use physical layer
            UtDataTypeId = (byte)244;
            LtDataTypeId = (byte)252;
            break;
        case (byte)0x04:    // UT: , LT: 
            UtDataTypeId = (byte)242;
            LtDataTypeId = (byte)250;
            break;
        case (byte)0xC0:    // UT: network&transport layer, LT: ?? layer
            UtDataTypeId = (byte)246;
            //LtDataType = (byte);
            break;
        case (byte)0x60:    // UT: facilities layer, LT: network&transport layer
            UtDataTypeId = (byte)244;
            LtDataTypeId = (byte)254;
            break;
        case (byte)0x24:    // UT: , LT: 
            UtDataTypeId = (byte)242;
            LtDataTypeId = (byte)252;
            break;
        case (byte)0xE0:    // UT: facilities layer, LT: facilities layer
            UtDataTypeId = (byte)244;
            //LtDataTypeId = (byte)252;
            break;
        case (byte)0x64:    // UT:  layer, LT: network&transport layer
            UtDataTypeId = (byte)242;
            LtDataTypeId = (byte)254;
            break;
        case (byte)0x74:    // UT:  layer, LT:  layer
            UtDataTypeId = (byte)242;
            //LtDataTypeId = (byte)254;
            break;
        case (byte)0x30:    // F-SAP, UT: facilities layer, LT: use physical layer
            UtDataTypeId = (byte)244;
            LtDataTypeId = (byte)252;
            break;
        case (byte)0x70:    // FSAP testing, UT: facilities layer, LT: facilities layer
            //UtDataTypeId = (byte)244;
            LtDataTypeId = (byte)254;
            break;
        case (byte)0xF0:    // not applicable UT: , LT: 
            //UtDataTypeId = (byte)244;
            //LtDataTypeId = (byte)254;
            break;
        /*
        case (byte)0x10:    // UT: management layer, LT: use physical layer
            UtDataTypeId = (byte)248;
            break;    */
        case (byte)0x01: // FIXME Sent by IICP ATS, not sure how to handle this. Check with HJ
            UtDataTypeId = (byte)240;
            LtDataTypeId = (byte)241;
            IicpTesting  = true;
            break;
        default: 
            UtDataTypeId = (byte)240; // YGA To be fixed
            LtDataTypeId = (byte)241;
            IicpTesting  = true;
            _logger.warning("Iicp.SetDataTypeIdentifiersForPorts, " + 
                    "TestConfigIICP = 0x" + Integer.toHexString(TestConfigIICP) + ", unexpected value");
            break;
        }
        _logger.info("Iicp.SetDataTypeIdentifiersForPorts, " + 
                    "TestConfigIICP = 0x" + Integer.toHexString(TestConfigIICP) +
                    ", UtDataType = " + Integer.toString((int)UtDataTypeId & 0xFF) +
                    ", LtDataType = " + Integer.toString((int)LtDataTypeId & 0xFF) +
                    ", CfDataType = " + Integer.toString((int)CfDataTypeId & 0xFF) +
                    ", IicpTesting = " + Boolean.toString(IicpTesting));
    }
    
    /**
     * 
     * @author AS
     *
     */
    final class DispatcherTriggerUt extends Thread implements IUdpImplEvent {
        
        public DispatcherTriggerUt(final String settings)
        {
            _logger.entering(this.toString(), "dispatcherTriggerUt", String.format("%s", settings));
            
            try {
                Pattern pattern = Pattern.compile("(\\w+)\\:(\\S+)\\:(\\d+)\\:(\\d+)"); // For Java regex test, see http://www.cis.upenn.edu/~matuszek/General/RegexTester/regex-tester.html
                Matcher matcher = pattern.matcher(settings);
                if (matcher.find() && (Integer.parseInt(matcher.group(3)) > 0) && (Integer.parseInt(matcher.group(4)) > 0)) {
                    UtActive = true;
                    _recvUtPacketBuffer = new LinkedList<byte[]>();
                    _recvUtPacketIdx = 0;
                    _processedUtPacketIdx = 0;
                    _udpUt = new UdpImpl();
                    _udpUt.initialize(matcher.group(2), Integer.parseInt(matcher.group(3)), Integer.parseInt(matcher.group(4)), this);
                } else {
                    //_logger.severe("Malformed parameters (2)");
                    //throw new RuntimeException("Malformed parameters (2)");
                    _logger.severe("UT not active");
                }
            } catch (UnknownHostException e) {
                e.printStackTrace();
            } catch (SocketException e) {
                e.printStackTrace();
            }
        } // End of Constructor
                
        /**
         * asynchronous called to handle incoming packets
         */
        @Override
        public void onReceive(byte[] packet) {
            ByteHelper.dump("DispatchTriggerUt.onReceive:", packet);
            synchronized(_recvUtPacketBuffer) {
                try {
                    _recvUtPacketBuffer.add(packet);    
                } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                    e.printStackTrace();
                }
                _recvUtPacketIdx++;
                if (_recvUtPacketIdx > _MaxBufferSize) _recvUtPacketIdx = 0;
            }
            if (debugBuffer) System.out.println("@@@@@ Ut.rx processedIdx=" + _processedUtPacketIdx + " - recvIdx=" + _recvUtPacketIdx + " - length=" + _recvUtPacketBuffer.size());
        }
        
        /**
         * dispatch received messages from the UT port
         * @param msg    - received message
         */
        public void DispatchMessage(byte[] msg)
        {
            if (IicpSimulate) {    // for debugging without SUT
                try {
                    if ((msg.length == 1) && (
                            (msg[0] == (byte)CfFntpInitializeTag) || 
                            (msg[0] == (byte)CfFsapInitializeTag) || 
                            (msg[0] == (byte)CfIicpInitializeTag) || 
                            (msg[0] == (byte)CfAtspInitializeTag) 
                    )) { // Config Tester Initialize
                     _udpCf.send(new byte[] {(byte)CfResultTag, (byte)0x01}); 
                    } else if ((msg.length == 3) && (msg[0] == (byte)0x90)) { // UtInitialize 
                     _udpUt.send(new byte[] {(byte)UtResultTag, (byte)0x01});            
                    } else if ((msg.length == 9) && (msg[2] == (byte)0x08)) { // create port 
                     _udpUt.send(new byte[] {(byte)0x00, (byte)0x05, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01});
                    }
                } catch (/*InterruptedException |*/ IOException e) {
                    e.printStackTrace();
                }
            } else {    // testing with real SUT using IICP port
                // check for special messages
                if ((msg.length == 1) && (
                        (msg[0] == (byte)CfFntpInitializeTag) || 
                        (msg[0] == (byte)CfFsapInitializeTag) || 
                        (msg[0] == (byte)CfIicpInitializeTag) || 
                        (msg[0] == (byte)CfAtspInitializeTag) 
                )) { // Config Tester Initialize
                    // special tester function to check the UDP connection, a UtResult message is expected
                    try {
                        _udpCf.send(new byte[] {    // send CfResult to upper tester port
                                (byte)CfResultTag,
                                (byte)0x01
                        });
                    } catch (/*InterruptedException |*/ IOException e) {
                        e.printStackTrace();
                    }
//                } else if ((msg.length == 3) && (msg[0] == (byte)0xFF)) {    // upper tester initialize
                } else if ((msg.length == 4) && ((msg[0] == UtFntpInitializeTag) || (msg[0] == UtFsapInitializeTag) || (msg[0] == UtIicpInitializeTag) || (msg[0] == UtAtspEventIndTag))) {    // upper tester initialize
                    // remove UtFntpInitialize tag
                    msg = ByteHelper.extract(msg, 1, msg.length - 1);
                    // save configuration/setup data for message routing
                    SetDataTypeIdentifiersForPorts(msg[2]);
                    pduCounterIicp += 2;
                    DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) CfDataTypeId, (byte)0, msg);
                } else { // UtxxxEvent
                    // Backup UtEbventTag which provides the message type
                    byte utEventTag = msg[0];
                    // remove UtxxxEvent tag
                    msg = ByteHelper.extract(msg, 1, msg.length - 1);
                    // IICP msg looks like <TestSystemScuId> <SutScuId> <pduCounter, even> <pduReq = 246>
                    if (IicpTesting) {
                        byte contentId = GetContentId();
                        if (utEventTag == UtIicpEventTag_mf_Request_request) { // Work-around to transmit a remote command
                            contentId = (byte)0xa2; // RefIICP.c-mfRreq
                        } else if (utEventTag == UtIicpEventTag_mn_Request_request) {
                            contentId = (byte)0xa4; // RefIICP.c-mnRreq
                        } else if (utEventTag == UtIicpEventTag_mi_Request_request) {
                            contentId = (byte)0xa6; // RefIICP.c-miRreq
                        } else if (utEventTag == UtIicpEventTag_mi_Get_request) {
                            contentId = (byte)0xa7; // RefIICP.c-miRget
                        } else if (utEventTag == UtIicpEventTag_mi_Set_request) {
                            contentId = (byte)0xa8; // RefIICP.c-miRset
                        } else if (utEventTag == UtIicpEventTag_mf_Command_confirm) {
                            contentId = (byte)0xa1; // RefIICP.c-mfRcmd
                        } else if (utEventTag == UtIicpEventTag_mn_Command_confirm) {
                            contentId = (byte)0xa3; // RefIICP.c-mnRcmd
                        } else if (utEventTag == UtIicpEventTag_mi_Command_confirm) {
                            contentId = (byte)0xa5; // RefIICP.c-miRcmd
                        }
                        pduCounterIicp += 2;
                        DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) UtDataTypeId, contentId, msg);
                    } else {
                        pduCounterIicp += 2;
                        DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) UtDataTypeId, (byte)0, msg);
                    }
                }
            }
        }     // End of Dispatch Message
    }    // end of DispatcherTriggerUt

    /**
     * 
     * @author AS
     *
     */
    final class DispatcherTriggerLt extends Thread implements IUdpImplEvent {
        
        public DispatcherTriggerLt(final String settings)
        {
            _logger.entering(this.toString(), "dispatcherTriggerLt", String.format("%s", settings));
            
            try {
                Pattern pattern = Pattern.compile("(\\w+)\\:(\\S+)\\:(\\d+)\\:(\\d+)"); // For Java regex test, see http://www.cis.upenn.edu/~matuszek/General/RegexTester/regex-tester.html
                Matcher matcher = pattern.matcher(settings);
                if (matcher.find() && (Integer.parseInt(matcher.group(3)) > 0) && (Integer.parseInt(matcher.group(4)) > 0)) {
                    LtActive = true;
                    _recvLtPacketBuffer = new LinkedList<byte[]>();
                    _recvLtPacketIdx = 0;
                    _processedLtPacketIdx = 0;
                    _udpLt = new UdpImpl();
                    _udpLt.initialize(matcher.group(2), Integer.parseInt(matcher.group(3)), Integer.parseInt(matcher.group(4)), this);
                } else {
                    //_logger.severe("Malformed parameters (2)");
                    //throw new RuntimeException("Malformed parameters (2)");
                    _logger.severe("LT not active");
                }
            } catch (UnknownHostException e) {
                e.printStackTrace();
            } catch (SocketException e) {
                e.printStackTrace();
            }
        } // End of Constructor
                
        
        /**
         * asynchronous called to handle incoming packets
         */
        @Override
        public void onReceive(byte[] packet) {
            ByteHelper.dump("DispatchTriggerLt.onReceive:", packet);
            synchronized(_recvLtPacketBuffer) {
                try {
                    _recvLtPacketBuffer.add(packet);    
                } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                    e.printStackTrace();
                }
                _recvLtPacketIdx++;
                if (_recvLtPacketIdx > _MaxBufferSize) _recvLtPacketIdx = 0;
            }
            if (debugBuffer) System.out.println("@@@@@ Lt.rx processedIdx=" + _processedLtPacketIdx + " - recvIdx=" + _recvLtPacketIdx + " - length=" + _recvLtPacketBuffer.size());
        }
        
        /**
         * dispatch received messages from the LT port
         * @param msg    - received message
         */
        public void DispatchMessage(byte[] msg)
        {
            if (IicpSimulate) {    // for debugging without SUT
                try {
                    if ((msg.length == 1) && (
                            (msg[0] == (byte)CfFntpInitializeTag) || 
                            (msg[0] == (byte)CfFsapInitializeTag) || 
                            (msg[0] == (byte)CfIicpInitializeTag) || 
                            (msg[0] == (byte)CfAtspInitializeTag) 
                    )) { // Config Tester Initialize
                     _udpCf.send(new byte[] {(byte)CfResultTag, (byte)0x01});            
                    } else if ((msg.length == 3) && (msg[0] == (byte)0x90)) { // UtInitialize 
                     _udpLt.send(new byte[] {(byte)UtResultTag, (byte)0x01}); 
                    } else if ((msg.length == 9) && (msg[2] == (byte)0x08)) { // create port 
                     _udpLt.send(new byte[] {(byte)0x00, (byte)0x05, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01});                    }
                } catch (/*InterruptedException |*/ IOException e) {
                    e.printStackTrace();
                }
            } else {    // testing with real SUT using IICP port
                // extract address fields from the message
                int  srcScuId   = (((int)msg[0] & 0xFF) << 8) + ((int)msg[1] & 0xFF);
                //int  destScuId  = (((int)msg[2] & 0xFF) << 8) + ((int)msg[3] & 0xFF);
                if (IicpTesting && (srcScuId != TestSystemScuId)) {
                    lastSimUtOutRqContentId = (byte)254;    // set to invalid
                    lastLtReqRef = msg[7];                // reqRef from last processed IIC-Request received on LT port    
                    try {
                        _udpIicp.send(msg);
                    } catch (/*InterruptedException |*/ IOException e) {
                        e.printStackTrace();
                    }
                
                } else {
                    // IICP msg looks like <TestSystemScuId> <SutScuId> <pduCounter, even> <pduReq = 252> <
                    pduCounterIicp += 2;
                    DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) LtDataTypeId, (byte)0, msg);
                }
            }
        }     // End of Dispatch Message
        
    }    // End of DispatcherTriggerLt

    /**
     * 
     * @author AS
     *
     */
    final class DispatcherTriggerCf extends Thread implements IUdpImplEvent {
        
        /**
         * 
         */
        public DispatcherTriggerCf(final String settings)
        {
            _logger.entering(this.toString(), "dispatcherTriggerCf", String.format("%s", settings));
            
            try {
                Pattern pattern = Pattern.compile("(\\w+)\\:(\\S+)\\:(\\d+)\\:(\\d+)"); // For Java regex test, see http://www.cis.upenn.edu/~matuszek/General/RegexTester/regex-tester.html
                Matcher matcher = pattern.matcher(settings);
                if (matcher.find() && (!matcher.group(2).equals("0.0.0.0") && Integer.parseInt(matcher.group(3)) > 0) && (Integer.parseInt(matcher.group(4)) > 0)) {
                    CfActive = true;
                    _recvCfPacketBuffer = new LinkedList<byte[]>();
                    _recvCfPacketIdx = 0;
                    _processedCfPacketIdx = 0;
                    _udpCf = new UdpImpl();
                    _udpCf.initialize(matcher.group(2), Integer.parseInt(matcher.group(3)), Integer.parseInt(matcher.group(4)), this);
                } else {
                    //_logger.severe("Malformed parameters (2)");
                    //throw new RuntimeException("Malformed parameters (2)");
                    _logger.severe("CT not active");
                }
            } catch (UnknownHostException e) {
                e.printStackTrace();
            } catch (SocketException e) {
                e.printStackTrace();
            }
        } // End of Constructor
                
        /**
         * asynchronous called to handle incoming packets
         */
        @Override
        public void onReceive(byte[] packet) {
            ByteHelper.dump("DispatchTriggerCf.onReceive:", packet);
            synchronized(_recvCfPacketBuffer) {
                try {
                    _recvCfPacketBuffer.add(packet);    
                } catch (IllegalStateException | ClassCastException | NullPointerException | IllegalArgumentException  e) {
                    e.printStackTrace();
                }
                _recvCfPacketIdx++;
                if (_recvCfPacketIdx > _MaxBufferSize) _recvCfPacketIdx = 0;
            }
            if (debugBuffer) System.out.println("@@@@@ Cf.rx processedIdx=" + _processedCfPacketIdx + " - recvIdx=" + _recvCfPacketIdx + " - length=" + _recvCfPacketBuffer.size());
        }

        /**
         * dispatch received messages from the CF port
         * @param msg    - received message
         */
        public void DispatchMessage(byte[] msg)
        {
            if (IicpSimulate) {    // for debugging without SUT
                try {
                    if ((msg.length == 1) && (
                            (msg[0] == (byte)CfFntpInitializeTag) || 
                            (msg[0] == (byte)CfFsapInitializeTag) || 
                            (msg[0] == (byte)CfIicpInitializeTag) || 
                            (msg[0] == (byte)CfAtspInitializeTag) 
                    )) { // Config Tester Initialize
                     _udpCf.send(new byte[] {(byte)CfResultTag, (byte)0x01}); 
                    } else if ((msg.length == 3) && (msg[0] == (byte)0x90)) { // UtInitialize 
                     _udpCf.send(new byte[] {(byte)UtResultTag, (byte)0x01}); 
                    } else if ((msg.length == 9) && (msg[2] == (byte)0x08)) { // create port 
                     _udpCf.send(new byte[] {(byte)0x00, (byte)0x05, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01, (byte)0x01}); 
                    }
                } catch (/*InterruptedException |*/ IOException e) {
                    e.printStackTrace();
                }
            } else {    // testing with real SUT using IICP port
                // check for special messages
                if ((msg.length == 1) && (
                        (msg[0] == (byte)CfFntpInitializeTag) || 
                        (msg[0] == (byte)CfFsapInitializeTag) || 
                        (msg[0] == (byte)CfIicpInitializeTag) || 
                        (msg[0] == (byte)CfAtspInitializeTag) 
                )) { // Config Tester Initialize
                    // special tester function to check the UDP connection, a UtResult message is expected
                    try {
                        _udpCf.send(new byte[] {    // send CfResult to CF port
                                (byte)CfResultTag,    //0x24,
                                (byte)0x01
                        });
                    } catch (/*InterruptedException |*/ IOException e) {
                        e.printStackTrace();
                    }
                } else if ((msg.length == 3) && (msg[0] == (byte)0x90)) {    // upper tester initialize
                    // save configuration/setup data for message routing
                    SetDataTypeIdentifiersForPorts(msg[2]);
                    pduCounterIicp += 2;
                       DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) CfDataTypeId, (byte)0, msg);
                } else {
                    // IICP msg looks like <TestSystemScuId> <SutScuId> <pduCounter, even> <pduReq = 9> <
                    pduCounterIicp += 2;
                    DispatcherTrigger.SendIicpPduRequest(TestSystemScuId, SutScuId, pduCounterIicp, (byte) CfDataTypeId, (byte)0, msg);
                }
            }
        }     // End of Dispatch Message
        
    }    // End of DispatcherTriggerCt

} // End of class DispatcherTrigger
