 /* packet-iicp.c
 * Routines for IICP packet dissection
 *
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>
#include <epan/asn1.h>
#include <epan/dissectors/packet-per.h>
#include <epan/prefs.h>

#include <stdio.h>
#include <string.h>

#include "packet-iicp.h"

#ifdef _MSC_VER
/* disable: "warning C4146: unary minus operator applied to unsigned type, result still unsigned" */
#pragma warning(disable:4146)
#endif

#define PNAME             "IICP Protocol (ISO 24102-4)"
#define PSNAME            "IICP (ISO 24102-4)"
#define PFNAME            "iicp"
#define IICP_PORT         1200    /* UDP port */
#define IICPP_PROTOCOL_ID 66

#include "packet-iicp-val.h"

/* Initialize the protocol and registered fields */
static int proto_iicp       = -1;
static int global_iicp_port = IICP_PORT;
static dissector_handle_t iicp_handle = NULL;

/* Dissector tables */
static dissector_table_t iicp_pdu_type_dissector_table;
static dissector_table_t iicp_pdu_request_dissector_table;
static dissector_table_t iicp_pdu_response_dissector_table;
static dissector_table_t iicp_mcmdrq_dissector_table;
static dissector_table_t iicp_mcmdrs_dissector_table;

static dissector_table_t msap_fa_sap_primitives_down_dissector_table;
static dissector_table_t msap_fa_sap_primitives_up_dissector_table;
static dissector_table_t msap_mf_command_request_dissector_table;
static dissector_table_t msap_mf_command_confirm_dissector_table;
static dissector_table_t msap_mf_request_request_dissector_table;
static dissector_table_t msap_mf_request_confirm_dissector_table;
static dissector_table_t msap_mn_command_request_dissector_table;
static dissector_table_t msap_mn_command_confirm_dissector_table;
static dissector_table_t msap_mn_request_request_dissector_table;
static dissector_table_t msap_mn_request_confirm_dissector_table;
static dissector_table_t msap_fwt_dissector_table;
static dissector_table_t msap_fwtnot_dissector_table;

static dissector_table_t fntp_nf_sap_primitives_down_dissector_table;
static dissector_table_t fntp_nf_sap_primitives_up_dissector_table;

static dissector_table_t llsap_in_sap_primitives_down_dissector_table;
static dissector_table_t llsap_in_sap_primitives_up_dissector_table;

static dissector_table_t llsap_acc_params_dissector_table;

/* Used to identify IICP/refIICP field 
   Set by 'dissect_iicp_pdu_type' function 
 */
static guint32 Ref_IICP; 
/* Used to identify IICPrequest/iicpReq field 
   Set by 'dissect_iicp_pdu_type_request' function 
*/
static guint32 Ref_IICPReq; 
/* Used to identify IICPresponse/iicpResp field 
   Set by 'dissect_iicp_pdu_type_response' function 
*/
static guint32 Ref_IICPResp; 
/* Used to identify McmdRq/mcmd field
   Set by 'dissect_iicp_mcmd_request' function 
 */
static guint32 Ref_McmdRq; 
/* Used to identify McmdRs/mcmd field
   Set by 'dissect_iicp_mcmd_response' function 
 */
static guint32 Ref_McmdRs; 
static guint32 Ref_COMMDOWN; 
static guint32 Ref_COMMUP; 
static guint32 Ref_MFCMDREQ; 
static guint32 Ref_MFCMDCONF; 
static guint32 Ref_MFREQREQ; 
static guint32 Ref_MFREQCONF; 
static guint32 Ref_MNCMDREQ; 
static guint32 Ref_MNCMDCONF; 
static guint32 Ref_MNREQREQ; 
static guint32 Ref_MNREQCONF; 
static guint32 Ref_FWT;
static guint32 Ref_FWTNOT;
static guint32 Ref_NFSAPDOWN; 
static guint32 Ref_NFSAPUP; 
static guint32 Ref_INSAPDOWN; 
static guint32 Ref_INSAPUP; 
static guint32 Ref_ACCPARAMS;

/* Dissector for IICP/refIICP field */
static int dissect_iicp_pdu_type(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
/* Dissector for IICPrequest/iicpReq field */
static int dissect_iicp_pdu_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
/* Dissector for IICPresponse/iicpResp field */
static int dissect_iicp_pdu_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
/* Dissector for McmdRq/mcmd field */
static int dissect_iicp_mcmd_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
/* Dissector for McmdRs/mcmd field */
static int dissect_iicp_mcmd_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);

static int dissect_msap_fa_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_fa_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mf_command_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mf_command_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mf_request_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mf_request_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mn_command_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mn_command_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mn_request_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_mn_request_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_fwt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_msap_fwtnot(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);

static int dissect_fntp_nf_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_fntp_nf_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);

static int dissect_llsap_in_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);
static int dissect_llsap_in_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);

static int dissect_llsap_acc_params(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_);

static int dissect_iicp_FNTPNPDU(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_);
static int dissect_iicp_SAM(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_);

void proto_reg_handoff_iicp(void);

#include "packet-iicp-hf.c"

/* Initialize the subtree pointers */
static int ett_iicp = -1;

#include "packet-iicp-ett.c"

#include "packet-iicp-fn.c"

static int dissect_iicp_pdu_type(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(iicp_pdu_type_dissector_table, Ref_IICP, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_iicp_pdu_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(iicp_pdu_request_dissector_table, Ref_IICPReq, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_iicp_pdu_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(iicp_pdu_response_dissector_table, Ref_IICPResp, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_iicp_mcmd_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(iicp_mcmdrq_dissector_table, Ref_McmdRq, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_iicp_mcmd_response(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(iicp_mcmdrs_dissector_table, Ref_McmdRs, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_fa_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_fa_sap_primitives_down_dissector_table, Ref_COMMDOWN, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_fa_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_fa_sap_primitives_up_dissector_table, Ref_COMMUP, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mf_command_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mf_command_request_dissector_table, Ref_MFCMDREQ, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mf_command_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mf_command_confirm_dissector_table, Ref_MFCMDCONF, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mf_request_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mf_request_request_dissector_table, Ref_MFREQREQ, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mf_request_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mf_request_confirm_dissector_table, Ref_MFREQCONF, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mn_command_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mn_command_request_dissector_table, Ref_MNCMDREQ, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mn_command_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mn_command_confirm_dissector_table, Ref_MNCMDCONF, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mn_request_request(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mn_request_request_dissector_table, Ref_MNREQREQ, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_msap_mn_request_confirm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_mn_request_confirm_dissector_table, Ref_MNREQCONF, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}
static int dissect_msap_fwt(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_fwt_dissector_table, Ref_FWT, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}
static int dissect_msap_fwtnot(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(msap_fwtnot_dissector_table, Ref_FWTNOT, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_fntp_nf_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(fntp_nf_sap_primitives_down_dissector_table, Ref_NFSAPDOWN, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}
static int dissect_fntp_nf_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(fntp_nf_sap_primitives_up_dissector_table, Ref_NFSAPUP, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_llsap_in_sap_primitives_down(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(llsap_in_sap_primitives_down_dissector_table, Ref_INSAPDOWN, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}
static int dissect_llsap_in_sap_primitives_up(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(llsap_in_sap_primitives_up_dissector_table, Ref_INSAPUP, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}

static int dissect_llsap_acc_params(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, void *data _U_)
{
  return (dissector_try_uint_new(llsap_acc_params_dissector_table, Ref_ACCPARAMS, tvb, pinfo, tree, FALSE, NULL)) ? tvb_length(tvb) : 0;
}



static void
dissect_iicp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  proto_item      *iicp_item = NULL;
  proto_tree      *iicp_tree = NULL;
  int             offset = 0;
  guint8          lsbPduCounter;

  /* make entry in the Protocol column on summary display */
  //if (check_col(pinfo->cinfo, COL_PROTOCOL))
    col_set_str(pinfo->cinfo, COL_PROTOCOL, PSNAME);
  //if (check_col(pinfo->cinfo, COL_INFO))
    col_set_str(pinfo->cinfo, COL_INFO, "Station-internal management communications (ISO 24102-4)");
  
  /* create the iicp protocol tree */
  if (tree) {
    iicp_item = proto_tree_add_item(tree, proto_iicp, tvb, 0, -1, FALSE);
    iicp_tree = proto_item_add_subtree(iicp_item, ett_iicp);

    tvb_get_guint8(tvb, offset++);
    tvb_get_guint8(tvb, offset++);

    tvb_get_guint8(tvb, offset++);
    tvb_get_guint8(tvb, offset++);

    tvb_get_guint8(tvb, offset++);
    lsbPduCounter = tvb_get_guint8(tvb, offset);
    if ((lsbPduCounter % 2) != 0) { // Odd value => response
      dissect_IIC_Response_PDU(tvb, pinfo, tree);	  
    } else {
      dissect_IIC_Request_PDU(tvb, pinfo, tree);
    }
  }
}
  
/*--- proto_register_iicp -------------------------------------------*/
void proto_register_iicp(void) {
  module_t *iicp_module;

  /* List of fields */
  static hf_register_info hf[] = {

#include "packet-iicp-hfarr.c"
  };

  /* List of subtrees */
  static gint *ett[] = {
                  &ett_iicp,
#include "packet-iicp-ettarr.c"
  };

  /* Register protocol */
  proto_iicp = proto_register_protocol(PNAME, PSNAME, PFNAME);
  /* Register dissector */
  register_dissector(PFNAME, dissect_iicp, proto_iicp);
  /* Register fields and subtrees */
  proto_register_field_array(proto_iicp, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
  /* Register CLASS dissector tables */
  iicp_pdu_type_dissector_table = register_dissector_table("iicp.proto", PNAME, FT_UINT32, BASE_DEC);

  iicp_pdu_request_dissector_table = register_dissector_table("iicp.req", "IICPpdu request", FT_UINT32, BASE_DEC);
  iicp_pdu_response_dissector_table = register_dissector_table("iicp.resp", "IICPpdu response", FT_UINT32, BASE_DEC);

  iicp_mcmdrq_dissector_table = register_dissector_table("iicp.mcmdrq", "MCMDRQ", FT_UINT32, BASE_DEC);
  iicp_mcmdrs_dissector_table = register_dissector_table("iicp.mcmdrs", "MCMDRS", FT_UINT32, BASE_DEC);

  msap_fa_sap_primitives_down_dissector_table = register_dissector_table("msap.commdown", "FA Sap Primitive Down", FT_UINT32, BASE_DEC);
  msap_fa_sap_primitives_up_dissector_table = register_dissector_table("msap.commup", "FA Sap Primitive Up", FT_UINT32, BASE_DEC);

  msap_mf_command_request_dissector_table = register_dissector_table("msap.mfcmdreq", "MF-Command-Req", FT_UINT32, BASE_DEC);
  msap_mf_command_confirm_dissector_table = register_dissector_table("msap.mfcmdconf", "MF-Command-Conf", FT_UINT32, BASE_DEC);

  msap_mf_request_request_dissector_table = register_dissector_table("msap.mfreqreq", "MF-Request-Req", FT_UINT32, BASE_DEC);
  msap_mf_request_confirm_dissector_table = register_dissector_table("msap.mfreqconf", "MF-Request-Conf", FT_UINT32, BASE_DEC);

  msap_mn_command_request_dissector_table = register_dissector_table("msap.mncmdreq", "MN-Command-Req", FT_UINT32, BASE_DEC);
  msap_mn_command_confirm_dissector_table = register_dissector_table("msap.mncmdconf", "MN-Command-Conf", FT_UINT32, BASE_DEC);

  msap_mn_request_request_dissector_table = register_dissector_table("msap.mnreqreq", "MN-Request-Req", FT_UINT32, BASE_DEC);
  msap_mn_request_confirm_dissector_table = register_dissector_table("msap.mnreqconf", "MF-Request-Conf", FT_UINT32, BASE_DEC);

  msap_fwt_dissector_table = register_dissector_table("msap.fwt", "FWT", FT_UINT32, BASE_DEC);
  msap_fwtnot_dissector_table = register_dissector_table("msap.fwtnot", "FWTNOT", FT_UINT32, BASE_DEC);

  fntp_nf_sap_primitives_down_dissector_table = register_dissector_table("fntp.nfsapdown", "NF Sap Primitive Down", FT_UINT32, BASE_DEC);
  fntp_nf_sap_primitives_up_dissector_table = register_dissector_table("fntp.nfsapup", "NF Sap Primitive Up", FT_UINT32, BASE_DEC);

  llsap_in_sap_primitives_down_dissector_table = register_dissector_table("llsap.insapdown", "IN Sap Primitive Down", FT_UINT32, BASE_DEC);
  llsap_in_sap_primitives_up_dissector_table = register_dissector_table("llsap.insapup", "IN Sap Primitive Up", FT_UINT32, BASE_DEC);

  llsap_acc_params_dissector_table = register_dissector_table("llsap.accparams", "Access Parameters", FT_UINT32, BASE_DEC);

  iicp_module = prefs_register_protocol(proto_iicp, proto_reg_handoff_iicp);

  prefs_register_uint_preference(iicp_module, "udp.port", "IICP UDP Port", "Set the default port for IICP messages (Default of 1200)", 10, &global_iicp_port);

}


/*--- proto_reg_handoff_iicp ---------------------------------------*/
void
proto_reg_handoff_iicp(void)
{
  static gboolean inited = FALSE;

  if( !inited ) {
    iicp_handle = create_dissector_handle(dissect_iicp, proto_iicp);
    inited = TRUE;
#include "packet-iicp-dis-tab.c"
  } else {
    dissector_delete_uint("udp.port", global_iicp_port, iicp_handle);
  }

  /* Set our port number for future use */
  dissector_add_uint("udp.port", global_iicp_port, iicp_handle);
}

///////////////////////////////////////////////////////////////////////////////
// Add not generated functions due to problem into CITSapplReq ASN.1 file

static int
dissect_CITSapplReq_ITSSappCPReqReg(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  return offset + 0;
}

static int
dissect_CITSapplReq_MediumCost(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  return offset + 0;
}

static int
dissect_CITSapplReq_ITSSappRxTxPortNo(tvbuff_t *tvb _U_, int offset _U_, asn1_ctx_t *actx _U_, proto_tree *tree _U_, int hf_index _U_) {
  return offset + 0;
}
